/**
 * MCS Media Computer Software
 * Copyright 2019 by Wilfried Klaas
 * Project: SPSEmulator
 * File: AbstractEmulator.java
 * EMail: W.Klaas@gmx.de
 * Created: 31.01.2019 wklaa_000
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 */
package de.mcs.tools.sps.emulator;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import de.mcs.tools.sps.SPSAssembler;
import de.mcs.tools.sps.emulator.model.WebSessionModel;
import de.mcs.tools.sps.emulator.model.WorkingModel;
import de.mcs.tools.sps.emulator.model.WorkingModel.WORKINGSTATE;
import de.mcs.tools.sps.exceptions.SyntaxError;
import de.mcs.tools.sps.mnemonic.Mnemonic;
import de.mcs.utils.Logger;

/**
 * @author wklaa_000
 *
 */
public abstract class AbstractEmulator implements Emulator {

  Logger log = Logger.getLogger(this.getClass());

  private WebSessionModel model;

  public AbstractEmulator(WebSessionModel model) {
    this.model = model;
  }

  /**
   * @return the model
   */
  public WebSessionModel getModel() {
    return model;
  }

  /**
   * starting the emulatiing process. Resetting all registers, addresses, stacks, outputs...
   * 
   * @return return true, if everything is ok.
   */
  @Override
  public boolean start() {
    WorkingModel work = model.getWork();
    work.setAddress((short) 0);
    work.setRaddress((short) 0);
    work.setPage((byte) 0);
    work.setRegisterA((byte) 0);
    work.setRegisterB((byte) 0);
    work.setRegisterC((byte) 0);
    work.setRegisterD((byte) 0);
    work.setRegisterE((byte) 0);
    work.setRegisterF((byte) 0);
    work.setStack(new ArrayList<>());
    work.setWorkingstate(WORKINGSTATE.EMULATE);

    model.getProgram().setActualLine(0);

    model.getOutput().setOutput1(false);
    model.getOutput().setOutput2(false);
    model.getOutput().setOutput3(false);
    model.getOutput().setOutput4(false);
    model.getOutput().setPwm1((byte) 0);
    model.getOutput().setPwm2((byte) 0);
    model.getOutput().setServo1((byte) 0);
    model.getOutput().setServo2((byte) 0);
    model.getOutput().setTone((byte) 0);

    return true;
  }

  /**
   * processing the next command from the program
   * 
   * @return return true, if everything is ok.
   */
  @Override
  public boolean next() {
    byte commandData = model.getProgram().getBin()[model.getProgram().getActualLine()];
    log.debug("new command 0x%02x", commandData);

    int command = (commandData & 0xF0);
    byte data = (byte) (commandData & 0x0F);

    switch (command) {
    case 0x00:
      doNop(data);
      break;
    case 0x10:
      doPort(data);
      break;
    case 0x20:
      doDelay(data);
      break;
    case 0x30:
      doJumpBack(data);
      break;
    case 0x40:
      doSetA(data);
      break;
    case 0x50:
      doIsA(data);
      break;
    case 0x60:
      doAIs(data);
      break;
    case 0x70:
      doCalc(data);
      break;
    case 0x80:
      doPage(data);
      break;
    case 0x90:
      doJump(data);
      break;
    case 0xA0:
      doCCount(data);
      break;
    case 0xB0:
      doDCount(data);
      break;
    case 0xC0:
      doSkip(data);
      break;
    case 0xD0:
      doCall(data);
      break;
    case 0xE0:
      doCallSub(data);
      break;
    case 0xF0:
      doByte(data);
      break;

    default:
      break;
    }
    model.getWork().setAddress((short) (model.getWork().getAddress() + 1));
    return false;
  }

  protected abstract void doSetA(byte data);

  protected abstract void doCalc(byte data);

  protected abstract void doCall(byte data);

  protected abstract void doCCount(byte data);

  protected abstract void doNop(byte data);

  protected abstract void doPort(byte data);

  protected abstract void doDelay(byte data);

  protected abstract void doJumpBack(byte data);

  protected abstract void doAIs(byte data);

  protected abstract void doIsA(byte data);

  protected abstract void doPage(byte data);

  protected abstract void doJump(byte data);

  protected abstract void doDCount(byte data);

  protected abstract void doSkip(byte data);

  protected abstract void doCallSub(byte data);

  protected abstract void doByte(byte data);

  @Override
  public boolean stop() {
    model.getWork().setWorkingstate(WORKINGSTATE.NN);
    return true;
  }

  @Override
  public boolean compile() throws SyntaxError, IOException {
    SPSAssembler spsAssembler = new SPSAssembler();
    List<String> source = Arrays.asList(model.getProgram().getSource());
    spsAssembler.doCompile(source);
    List<Mnemonic> mnemonics = spsAssembler.getMnemonics();
    List<Integer> lineNumbers = spsAssembler.getLines();
    byte[] bin = new byte[mnemonics.size()];
    int[] bin2SrcLine = new int[mnemonics.size()];
    for (int i = 0; i < mnemonics.size(); i++) {
      Mnemonic mnemonic = mnemonics.get(i);
      bin[i] = (byte) mnemonic.getByte();
      bin2SrcLine[i] = lineNumbers.get(i);
    }
    model.getProgram().setBin(bin);
    model.getProgram().setBin2SrcLine(bin2SrcLine);

    return false;
  }

}
