package de.mcs.utils;
/**
 * 
 */

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * This implements a iterable list of unique {@link Macro}.
 * 
 * @author w.klaas
 */
public class Macros implements Iterable<Macro> {

  private List<Macro> macroTable;

  /**
   * Default constructor for the macro list
   */
  public Macros() {
    this.macroTable = new ArrayList<>();
  }

  /**
   * adding a new macro definition to this list. If the name is already present, only the value will be changed.
   * 
   * @param name
   *          name of the new macro
   * @param value
   *          substitution value of the macro
   */
  public void addMacro(String name, String value) {
    for (Iterator<Macro> iterator = macroTable.iterator(); iterator.hasNext();) {
      Macro macro = (Macro) iterator.next();
      if (macro.getMacroName().equals(name)) {
        macro.setValue(value);
        return;
      }
    }
    macroTable.add(new Macro(name, value));
  }

  /**
   * adding a new macro definition to this list. If the name is already present, only the value will be changed.
   * 
   * @param macro
   *          the macro to be added
   */
  public void addMacro(Macro macro) {
    for (Iterator<Macro> iterator = macroTable.iterator(); iterator.hasNext();) {
      Macro myMacro = (Macro) iterator.next();
      if (myMacro.getMacroName().equals(macro.getMacroName())) {
        myMacro.setValue(macro.getValue());
        return;
      }
    }
    macroTable.add(macro);
  }

  /**
   * @return iterator over this list of macros
   */
  public Iterator<Macro> iterator() {
    return macroTable.iterator();
  }

  /**
   * substitue a string with macros. All Macros should be name in the string as {macroname}.
   * 
   * @param text
   *          the original string with the macro definitions in
   * @return the substituted string
   */
  public String substitute(final String text) {
    StringBuilder stringBuilder = new StringBuilder(text);
    for (Macro macro : this) {
      int index = stringBuilder.indexOf(macro.getMacroDefinition());
      if (index != -1) {
        stringBuilder.replace(index, index + macro.getMacroDefinition().length(), macro.getValue());
      }
    }
    return stringBuilder.toString();
  }

  /**
   * getting a single macro from the list of macros
   * 
   * @param name
   *          the name of the macros to get
   * @return the desired macro or <code>null</code> if the macro is not present.
   */
  public Macro getMacro(String name) {
    for (Iterator<Macro> iterator = macroTable.iterator(); iterator.hasNext();) {
      Macro macro = (Macro) iterator.next();
      if (macro.getMacroName().equals(name)) {
        return macro;
      }
    }
    return null;
  }

  /*
   * (non-Javadoc)
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    StringBuilder b = new StringBuilder();
    for (Iterator<Macro> iterator = macroTable.iterator(); iterator.hasNext();) {
      Macro macro = (Macro) iterator.next();
      b.append(macro.toString());
      b.append(',');
    }
    return String.format("[%s]", b.substring(0, b.length() - 1));
  }

}
