/**
 * MCS Media Computer Software
 * Copyright 2018 by Wilfried Klaas
 * Project: SPSEmulator
 * File: Not.java
 * EMail: W.Klaas@gmx.de
 * Created: 25.11.2018 wklaa_000
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 */
package de.mcs.tools.sps.mnemonic;

import java.util.EnumSet;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;

import de.mcs.tools.sps.exceptions.IllegalArgument;
import de.mcs.tools.sps.exceptions.SyntaxError;

//@formatter:off
/**
 * BLDA ADC#x: byte analog input #x to register A. x should be 1 or 2
 * BLDA RC#x: byte rc receiver input #x to register A. x should be 1 or 2
 * @author wklaa_000
 *
 */
//@formatter:on
public class BLDA extends AbstractMnemonic implements Mnemonic {

  byte value;

  enum INPUTSOURCE {
    ANALOG, RC
  };

  INPUTSOURCE inputSource;

  public BLDA(String line) throws SyntaxError {
    super(line);
    value = 0;
  }

  @Override
  public void checkArgument() throws SyntaxError {
    if (StringUtils.isEmpty(getArgument())) {
      throw new SyntaxError(getLineNumber(),
          String.format("missing argument for %s.", this.getClass().getSimpleName()));
    }

    String inputString = getArgument().toUpperCase();
    if (inputString.startsWith("ADC")) {
      inputSource = INPUTSOURCE.ANALOG;
      int myValue = 0;
      try {
        myValue = Integer.parseInt(inputString.substring(3));
      } catch (NumberFormatException e) {
      }
      if ((myValue < 1) || (myValue > 2)) {
        throw new IllegalArgument(getLineNumber(), String.format("argument %s is not in range ADC1..2 for %s.",
            getArgument(), this.getClass().getSimpleName()));
      }
      value = (byte) ((myValue - 1) & 0xFF);
    } else if (inputString.startsWith("RC")) {
      inputSource = INPUTSOURCE.RC;
      int myValue = 0;
      try {
        myValue = Integer.parseInt(inputString.substring(2));
      } catch (NumberFormatException e) {
      }
      if ((myValue < 1) || (myValue > 2)) {
        throw new IllegalArgument(getLineNumber(), String.format("argument %s is not in range RC1..2 for %s.",
            getArgument(), this.getClass().getSimpleName()));
      }
      value = (byte) ((myValue - 1) & 0xFF);
    } else {
      throw new SyntaxError(getLineNumber(),
          String.format("unknown argument %s for %s.", this.getArgument(), this.getClass().getSimpleName()));
    }
  }

  @Override
  public int getByte() {
    switch (inputSource) {
    case ANALOG:
      return 0xF0 + value;
    case RC:
      return 0xF2 + value;
    }
    return 0x00;
  }

  @Override
  public boolean isUsingPage() {
    return false;
  }

  @Override
  public boolean hasArgument() {
    return true;
  }

  @Override
  public Set<HARDWARE> allowedHardware() {
    return EnumSet.of(HARDWARE.ARDUINOSPS, HARDWARE.TINYSPS);
  }

}
